module FFJPL.FeatureModel
  ( Formula(..), FeatureModel,
    sometimes, never, alwaysOR, alwaysGroup,
    solve
  )
where {
	import FFJPL.Definitions;
	
	import Debug.Trace ( trace );
	
	import Control.Exception ( evaluate );
	import Data.IORef;
	import Data.Map ( Map );
	import qualified Data.Map as Map;
	import System.IO.Unsafe ( unsafePerformIO );

	data Formula = Tru | Fls | FVar Feature | And [Formula] | Or [Formula] | Not Formula | Impl Formula Formula deriving (Eq, Ord, Show);

	type FeatureModel = ([Feature],[Formula]);

	sometimes :: FeatureModel -> [Feature] -> Feature -> Bool; -- feature model interface
	sometimes (fs,fa) omegas psi = not (solve (Not (And (fa ++ (getFVars omegas) ++ [(FVar psi)]))));

	always :: FeatureModel -> [Feature] -> Feature -> Bool; -- feature model interface
	always (fs,fa) omegas psi = solve (Impl (And fa) (Impl (And (getFVars omegas)) (FVar psi)));

	alwaysOR :: FeatureModel -> [Feature] -> [Feature] -> Bool;
	alwaysOR (fs,fa) omegas psis = solve (Impl (And fa) (Impl (And (getFVars omegas)) (Or (getFVars psis))));

	never :: FeatureModel-> [Feature] -> Feature -> Bool; -- feature model interface
	never fm omegas psi = not (sometimes fm omegas psi);

	alwaysGroup :: FeatureModel -> [Feature] -> Feature -> [Feature];
	alwaysGroup fm omegas psi = if (always fm omegas psi) 
	                               then [psi]
	                               else if (alwaysOR fm omegas (psi:(neverGroup fm omegas psi)))
	                                    then (psi:(neverGroup fm omegas psi))
	                                    else [];
	neverGroup :: FeatureModel -> [Feature] -> Feature -> [Feature];
	neverGroup (fs,fa) omegas psi = neverGroupIter (fs,fa) fs omegas psi;

	neverGroupIter :: FeatureModel -> [Feature] -> [Feature] -> Feature -> [Feature];
	neverGroupIter fm (f:rest) omegas psi = if (never fm (f:omegas) psi) 
	                                    then (neverGroupIter fm rest omegas psi) ++ [f]
	                                    else (neverGroupIter fm rest omegas psi);
	neverGroupIter fm [] omegas psi = [];
	
	getFVars :: [Feature] -> [Formula];
	getFVars (f:fs) = (FVar f):(getFVars fs);
	getFVars [] = [];

    --
    -- A cache for the results of solveWithoutCache.
    --
    {-# NOTINLINE solveCache #-};
    solveCache :: IORef (Map Formula Bool);
    solveCache = unsafePerformIO $ newIORef Map.empty;
    
    solve :: Formula -> Bool;
    solve f = unsafePerformIO $
      do { -- Here, we need an ugly hack to avoid deadlocks which can occur
           -- due to the laziness in the updates of solveCache and
           -- the locking implied by atomicModifyIORef.
           -- We must make sure that solveWithCache (which is executed
           -- with the lock on solveCache held) does not evaluate f
           -- because the value of f can depend on a former invocation
           -- of solve and then evaluating f causes another instance
           -- of solve to be evaluated (and a deadlock occurs).
           -- So, we force the evaluation of f.
           -- (Note,   evaluate f  is not enough because of WHNF.)
           --
           evaluate (length $ show f);
           atomicModifyIORef solveCache (solveWithCache f);
         };

    solveWithCache :: Formula -> Map Formula Bool -> (Map Formula Bool, Bool);
    solveWithCache f cache =
      case Map.lookup f cache of {
        Just b  ->  (cache, b);
        Nothing ->  (newCache, res);
      }
      where {
        newCache = Map.insert f res cache;
        res = solveWithoutCache f;
      };

    --
    -- Return True if the given formula is a tautology
    -- (i.e., the formula is true for all possible
    -- assignments to the variables).
    --
	solveWithoutCache :: Formula -> Bool;
	solveWithoutCache = solve1 . simplifyTF; -- . (\s -> trace (show s) s);
	
	--
	-- Rewrite f to
	--   f[x:=Tru] and f[x:=Fls]
	-- if there is a variable x in f.
	-- f is a tautology  iff  the above term is a tautology.
	-- If f is variable-free, return the boolen value of f.
	---
	solve1 :: Formula -> Bool;
	solve1 f = case (findFVar f) of {
		Just v  -> solve1 $ simplifyTF $ And [(assign v Tru f),(assign v Fls f)];
		Nothing ->  boolean f;
	}; 
	
	findFVar :: Formula -> Maybe Feature;
	
	findFVar (FVar n) = Just n;
	
	findFVar (And (f:fs)) = case (findFVar f) of {
		Just v -> Just v;
		Nothing -> findFVar (And fs);
	};
	findFVar (And []) = Nothing;

	findFVar (Or (f:fs)) = case (findFVar f) of {
		Just v -> Just v;
		Nothing -> findFVar (Or fs);
	};
	findFVar (Or []) = Nothing;
	
	findFVar (Not f) = findFVar f;
	
	findFVar (Impl f g) = case (findFVar f) of {
		Just v -> Just v;
		Nothing -> findFVar g;
	};

	findFVar Tru = Nothing;
	
	findFVar Fls = Nothing;
	
	assign :: Feature -> Formula -> Formula -> Formula;
	assign n c (FVar s) | (n == s) = c
                        | otherwise = (FVar s);
    assign n c (And fs) =  (And (map (assign n c) fs));
    assign n c (Or fs) =  (Or (map (assign n c) fs));
    assign n c (Impl f1 f2) = (Impl (assign n c f1) (assign n c f2));
    assign n c (Not f) = (Not (assign n c f));
    assign n c (Tru) = Tru;
   	assign n c (Fls) = Fls;

   	--
   	-- Evaluate a formula without variables.
   	--
   	boolean :: Formula -> Bool;
   	boolean (And fs) = (foldl (&&) True (map boolean fs)); 
   	boolean (Or fs) = (foldl (||) False (map boolean fs));
   	boolean (Impl f1 f2) = ((not (boolean f1)) || (boolean f2));
   	boolean (Not f) = (not (boolean f));
   	boolean (Tru) = True;
   	boolean (Fls) = False;
   	boolean f = error ("boolean: tried to evaluate " ++ show f);
    
    --
    -- Perform basic formula simplification.
    -- Replace conjunctions/disjunction/implications
    -- by their truth value if the it can be determined
    -- from the partially availabe truth values of the
    -- operands.
    --
    simplifyTF :: Formula -> Formula;
    simplifyTF f@(FVar _) = f;
    simplifyTF Tru        = Tru;
    simplifyTF Fls        = Fls;
    simplifyTF (Not f)    = case simplifyTF f of {
                              Tru -> Fls;
                              Fls -> Tru;
                              g   -> Not g;
                            };
    simplifyTF (And fs)
        | any (==Fls) gs  =  Fls
        | otherwise       =  if null gs then Tru else And gs
        where {
          gs = [g | g <- map simplifyTF fs, g /= Tru];
        };
    simplifyTF (Or fs)
        | any (==Tru) gs  =  Tru
        | otherwise       =  if null gs then Fls else Or gs
        where {
          gs = [g | g <- map simplifyTF fs, g /= Fls];
        };
    simplifyTF (Impl f1 f2) = case (simplifyTF f1, simplifyTF f2) of {
                                (Fls, _  ) -> Tru;
                                (_  , Tru) -> Tru;
                                (Tru, Fls) -> Fls;
                                (g1 , g2 ) -> Impl g1 g2;
                              };
}
