Parser for Featherweight Java
(C) Andreas Abel 2002

> {
> {-# OPTIONS -fno-warn-overlapping-patterns #-}
>
> module FJ.Parser where
>
> import Char;
> }


First thing to declare is the name of your parser,
and the type of the tokens the parser reads.

> %name parser
> %tokentype { Token }


The parser will be of type [Token] -> ?, where ? is determined by the
production rules.  Now we declare all the possible tokens:

> %token 
>       class           { TokenClass }
>       extends         { TokenExtends }
>       object          { TokenObject }
>       this            { TokenThis }
>       new             { TokenNew }
>       return          { TokenReturn }
>       super           { TokenSuper }
>	id		{ TokenID $$ }
>       '.'             { TokenDot }
>       ','             { TokenComma }
>       ';'             { TokenSemicolon }
>	'='		{ TokenEq }
>	'('		{ TokenLParen }
>	')'		{ TokenRParen }
>       '{'             { TokenLBrace }
>       '}'             { TokenRBrace }


Like yacc, we include %% here, for no real reason.

> %%

Statements.

We parse a (possibly empty) list of class (Class) and variable (Def)
definitions.

    Stats :                no more statements
          | Class Stats    "class c extends d { ... }" and more
          | Def Stats      "c x = t;" and more

Since the parser generator does not support right recursive rules as
this one, we transform it into a left recursive rule.  This parses the
statements in the wrong order, from last to first.  We fix this by
finally reversing the list of statements.  (Read more in the Happy
Tutorial by Simon Marlow and Andy Gill.)

> Stats :: { [PStat] }
> Stats : Stats1                    { reverse $1 }
>
> Stats1 :: { [PStat] }
> Stats1 : Stats1 Class             { $2 : $1}
>        | Stats1 Def               { $2 : $1}
>        |                          { [] }

Class production rules.

A class consists of zero or more fields, a constructor and zero or
more methods.

    Class  : class id extends CID '{' Fields Constr Methods '}'

    Fields :             
           | CID id; Fields

    Constr : CID (...) { super (...); FInits }

A class identifier CID is either "Object" or an identifier. As Java, we permit both upper and lower case identifiers for classes.

> CID   :: { PCID }
> CID   : object                    { PObject }
>       | id                        { PCID $1 }

Since both field and constructor declarations start with an identifier,
we need cannot decide right from the start which of these declarations
we are parsing.  We need to see another token.  Is it an identifier,
then we are parsing a field declaration, is it an opening parenthesis,
we are parsing the constructor definition.  Hence, we introduce the
non-terminal FC which denotes "a possibly empty list of field
declarations plus finally the constructor code".

> Class :: { PStat }
> Class : class id extends CID '{' FC Methods '}'   { PClass $2 $4 $6 $7 } 
>
> FC    :: { PFC }
> FC    : CID id ';' FC             { PField (PDecl $1 $2) $4 }
>       | id '(' Declo ')' '{' super '(' IDo ')' ';' FInits '}'  { PConstr $1 $3 $8 $11 }

Decls is a non empty list of "CID ID" declarations as used for
constructor and method argument lists.  Declo is a possibly empty list
of such declarations (the "o" for "optional).

> Declo :: { [PDecl] }
> Declo :                           { [] }
>       | Decls                     { reverse $1 }
>
> Decls :: { [PDecl] }
> Decls : Decls ',' CID id          { PDecl $3 $4 : $1 }
>       | CID id                    { [PDecl $1 $2] }

IDs and IDo is similiar to Decls and Declo, but here it is just a list
of identifiers.

> IDo   :: { [String] }
> IDo   :                           { [] }
>       | IDs                       { reverse $1 }
>
> IDs   :: { [String] }
> IDs   : IDs ',' id                { $3 : $1 }
>       | id                        { [$1] }

Field initializations in the constructor body have the fixed form
"this.ID = ID;".  FInits is a possibly empty list of initializations.

> FInits :: { [(String,String)] }
> FInits : FInit1                   { reverse $1 }
>
> FInit1 :: { [(String,String)] }
> FInit1 : FInit1 this '.' id '=' id ';'     { ($4,$6) : $1 }
>        |                                   { [] }

Method definitions.

    Methods :
            | CID id (Declo) { return Term; } Methods

> Methods :: { [PMethod] }
> Methods : Method1                 { reverse $1 }
>
> Method1 :: { [PMethod] }
> Method1 : Method1 Method          { $2 : $1 }
>         |                         { [] }
>
> Method :: { PMethod }
> Method : CID id '(' Declo ')' '{' return Term ';' '}'  { PMethod (PDecl $1 $2) $4 $8 } 

Term production rules.

    Term : Var              Variable 'this' or id
         | Term Elims       Elimination: field lookup or method invokation
         | (CID) Term       Cast
         | (Term)           Parentheses to group terms and resolve ambiguities
         | new CID (Terms)  Object construction

The main issue here is to resolve the ambiguity between terms in
parentheses (Term) and casts (CID) Term.  Since there is no reason to
put a single variable in parentheses, we interprete "(ID)" as cast.
In practise, we need to distiguish terms which consist of a single
variable from more complex terms (Term1).

> Term  :: { PTerm }
> Term  : Var                       { PVar $1 }
>       | Term1                     { $1 }
>
> Term1 :: { PTerm }
> Term1 : Var '.' Elim Elims        { PElim (PVar $1) ($3 : $4) }
>       | '(' CID ')' Term          { PCast $2 $4}
>       | '(' Term1 ')' Elims       { PElim $2 $4 }  
>       | new CID '(' Termo ')'     { PNew $2 $4 }
>
> Var   :: { PVar }
> Var   : this                      { PThis }
>       | id                        { PVID $1 }

An elimination Elim is either a field lookup ("Get") or a method
invocation ("Call").

> Elims :: { [PElim] }
> Elims : Elims1                    { reverse $1 }
>
> Elims1:: { [PElim] }
> Elims1: Elims '.' Elim            { $3 : $1 }
>       |                           { [] }
>
> Elim  :: { PElim }
> Elim  : id '(' Termo ')'          { PCall $1 $3 }
>       | id                        { PGet $1 }

Term lists.

> Termo :: { [PTerm] }
> Termo :                           { [] }
>       | Terms                     { reverse $1 }
>
> Terms :: { [PTerm] }
> Terms : Terms ',' Term            { $3 : $1 }
>       | Term                      { [$1] }

Variable definitions. 

> Def   :: { PStat }
> Def   : CID id '=' Term ';'       { PDef $1 $2 $4 }

> {

Pre-terms as parser result.

> data PVar    = PThis 
>       | PVID String
>	 deriving Show
> data PCID    = PObject 
>       | PCID String
>	 deriving Show
> data PTerm   = PVar PVar
>       | PElim PTerm [PElim]
>       | PCast PCID PTerm
>       | PNew PCID [PTerm]
>	 deriving Show
> data PElim   = PGet String
>       | PCall String [PTerm]
>	 deriving Show

Pre-statements.

> data PDecl   = PDecl PCID String
>	 deriving Show
> data PMethod = PMethod PDecl [PDecl] PTerm
>	 deriving Show
> data PFC     = PConstr String [PDecl] [String] [(String,String)]
>       | PField PDecl PFC
>	 deriving Show
> data PStat   = PClass String PCID PFC [PMethod]
>       | PDef PCID String PTerm 
>	 deriving Show

Parser error.

All parsers must declair this function, which is called when an error
is detected.  Note that currently we do no error recovery.

> happyError :: [Token] -> a
> happyError toks = error ("Parse error" ++ show toks ++ "\n")

The lexer.

> data Token = TokenClass
>     | TokenExtends
>     | TokenObject
>     | TokenThis
>     | TokenNew
>     | TokenReturn
>     | TokenSuper
>     | TokenID String
>     | TokenDot
>     | TokenComma
>     | TokenSemicolon
>     | TokenEq
>     | TokenLParen
>     | TokenRParen
>     | TokenLBrace
>     | TokenRBrace
>       deriving Show

> lexer :: String -> [Token]
> lexer [] = []
> lexer (c:cs) 
>	| isSpace c = lexer cs
> 	| isAlpha c = lexID (c:cs)
> lexer ('.':cs) = TokenDot       : lexer cs
> lexer (',':cs) = TokenComma     : lexer cs
> lexer (';':cs) = TokenSemicolon : lexer cs
> lexer ('=':cs) = TokenEq        : lexer cs
> lexer ('(':cs) = TokenLParen    : lexer cs
> lexer (')':cs) = TokenRParen    : lexer cs
> lexer ('{':cs) = TokenLBrace    : lexer cs
> lexer ('}':cs) = TokenRBrace    : lexer cs
> lexer ('/':'/':cs) = lexer (dropWhile (not . (==) '\n') cs)
> lexer ('/':'*':cs) = lexComment cs

> lexComment [] = error "unclosed comment at end of input"
> lexComment ('*':'/':cs) = lexer cs
> lexComment (_:cs) = lexComment cs

> lexID cs =
>    case span isAlphaNum cs of
>	("class",   rest) -> TokenClass   : lexer rest
>	("extends", rest) -> TokenExtends : lexer rest
>	("Object",  rest) -> TokenObject  : lexer rest
>	("super",   rest) -> TokenSuper   : lexer rest
>	("this",    rest) -> TokenThis    : lexer rest
>	("return",  rest) -> TokenReturn  : lexer rest
>	("new",     rest) -> TokenNew     : lexer rest
>	(id,        rest) -> TokenID id   : lexer rest

main = print . parser . lexer
>
>
End of file.

> }
